@@ -49,7 +49,7 @@ void checkisvar(int aff, const char *func, int *var, int *t)
       *t=tree[*var].y;
       *var=tree[*var].x;
       if (tree[*var].f!=Fentry)
-	die(aff,"incorrect syntax for %s",func);
+        die(aff,"incorrect syntax for %s",func);
       break;
     default:
       die(aff,"incorrect syntax for %s",func);
@@ -62,16 +62,16 @@ int guesstype(int n)
   {
   case Fsmall:
     return Gsmall;
-  case Fgnil:
+  case Fnoarg:
     return Gvoid;
   case Ftag:
     return tree[n].y;
-  case Fmat: case Fvec: 
+  case Fmat: case Fvec:
     return Gvec;
   case Fconst:
     switch(value[tree[n].x].type)
     {
-    case CSTsmall:      
+    case CSTsmall:
       return Gsmall;
     case CSTint:
       return Gint;
@@ -83,7 +83,10 @@ int guesstype(int n)
     }
   case Fentry:
     return vartype(ctxstack[getvarerr(n)]);
-  default:
+  case Ffunction:
+    if (!strcmp(entryname(n),"length"))
+      return Glg;
+  default: /*FALL THROUGH*/
     return Gnotype;
   }
 }
@@ -114,44 +117,43 @@ int gengerepile(int bl)
   int arg[4];
   block[bl].gc|=(1<<GCneeded);
   arg[0]=newnode(Fentry,newentry("btop"),-1);
-  pushvar(arg[0],1<<Cuser,Gpari_sp,newcall("_avma",-1));
+  pushvar(arg[0],0,Gpari_sp,newcall("_avma",-1));
   arg[1]=newnode(Fentry,newentry("st_lim"),-1);
-  pushvar(arg[1],1<<Cuser,Gpari_sp,
+  pushvar(arg[1],0,Gpari_sp,
       newcall("_stack_lim",newnode(Flistarg,arg[0],newsmall(1))));
-  arg[2]=(FC_gerepileall<0)?newnode(Fentry,newentry("bptr"),-1):GNOARG;
-  arg[3]=newsmall(bl);
-  return listtoseq(arg,4);
-}
-
-void gengerepileend(int bl)
-{
-  int gptr=newnode(Fentry,newentry("bptr"),-1);
-  int blo=newsmall(bl);
-  if (FC_gerepileall<0)
-    pushvar(gptr,1<<Cuser,Ggptr,newcall("_gerepilelist",blo));
+  arg[2]=newsmall(bl);
+  return listtoseq(arg,3);
 }
 
 int newgetg(int v, int l, const char *t)
 {
-  return newcall("_cgetg", newnode(Flistarg, newnode(Flistarg, 
+  return newcall("_cgetg", newnode(Flistarg, newnode(Flistarg,
           v, l), newstringnode(t,-1)));
 }
 
-int genblockinit(int seq, int start, int narg, int isglobal)
+void makesubblock_nolabel(int n)
 {
-  int i;
-  for(i=narg-1;i>=0;i--)
-  {
-    ctxvar *v=ctxstack+start+i;
-    int val=v->initval;
-    if (val>=0 && (isglobal || tree[val].f!=Fsmall))
-    {
-      int n=newnode(Faffect,newleaf(v->node),val);
-      seq=addseqleft(n,seq);
-      v->initval=-1;
-    } 
-  }
-  return seq;
+  int savx=s_ctx.n;
+  int bl3=newblock();
+  genblock_nolabel(n,-1);
+  makeblock(bl3,n,newleaf(n),-1,savx);
+}
+
+void makesubblock(int n)
+{
+  int savx=s_ctx.n;
+  int bl3=newblock();
+  genblock(n,-1);
+  makeblock(bl3,n,newleaf(n),-1,savx);
+}
+
+Clabel_t makesubblock_label(int n)
+{
+  int savx=s_ctx.n;
+  int bl3=newblock();
+  Clabel_t lab=genblock_label(n,-1);
+  makeblock(bl3,n,newleaf(n),-1,savx);
+  return lab;
 }
 
 /*
@@ -163,16 +165,15 @@ are:
   bseq=geninsertvar(arg,var)
   genblock(bseq,p)
   seq=addseqright(seq,newcall("_gerepilemany",gpm));
-  gengerepileend(bl2);
   savx=s_ctx.n
-  
+
   savx=s_ctx.n start the recording of var in the block
   pushvar/newdecl(var) must be before geninsert(arg,var) obviously,
 
   if you use geninsertvar to store the value a a node, then
   geninsertvar must be called before genblock and genblock must be
   called on the geninsertvar node, with p=-1 since the return value is really
-  ignored.  
+  ignored.
   G.N.^H^H^H^HB.A. (well it still lacks 20 lines more to really be G.N.)*/
 void genblockfuncspec(int n, int p, gpfunc *gp)
 {
@@ -180,12 +181,12 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
   int nb;
   int y=tree[n].y;
   int binf,bsup,seq;
-  int var,vinf,vsup,vprime;
+  int var=-1,vinf,vsup,vprime;
   int bseq=-1,aseq=-1,ret=-1;
-  int tv,vret=-1;
+  int tv;
   int gpm=-1,bl1,bl2;
   int savx;
-  label_t lab;
+  Clabel_t lab;
   switch(gp->spec)
   {
     case GPaddhelp:
@@ -205,7 +206,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     }
     break;
   case GPgrandO:
-    if (tree[y].f==Fentryfunc && tree[y].x==OPpow)
+    if (tree[y].f==Ffunction && tree[y].x==OPpow)
       tree[y]=tree[tree[y].y];
     tree[n].x=newentry("O(_^_)");
     genblock(y,n);
@@ -226,9 +227,9 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     {
       /*if(1,...) expr are often used to create block. We honor it here*/
       if (p>=0 && tree[p].f!=Fseq)
-	newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&ret);
+        newdecl((1<<Cauto)|(1<<Cconst),Gvoid,GNIL,&ret);
       arg[1]=geninsertvar(arg[1],ret);
-      genblock(arg[1],-1);
+      makesubblock(arg[1]);
       aseq=arg[1];
     }
     else
@@ -236,17 +237,17 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (p>=0 && tree[p].f!=Fseq)
       {
         if(nb==3)
-	  newdecl((1<<Cauto)|(1<<Cimmutable),Gempty,-1,&ret);
+          newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
         else
-	  newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
+          newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
       }
       genblock(arg[0],n);
       arg[1]=geninsertvar(arg[1],ret);
-      genblock(arg[1],-1);
+      makesubblock(arg[1]);
       if(nb==3)
       {
         arg[2]=geninsertvar(arg[2],ret==-1?-1:newleaf(ret));
-        genblock(arg[2],-1);
+        makesubblock(arg[2]);
       }
       aseq=newnode(tree[n].f,tree[n].x,listtoseq(arg,nb));
     }
@@ -276,10 +277,10 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       gp->proto.cname=entryname(d);
       if (nb==4 && arg[3]!=GNOARG)
       {
-	d=detag(arg[3]);
-	if (!is_const(d,CSTstr))
-	  die(n,"Incorrect library name in install");
-	gp->proto.origin=entryname(d);
+        d=detag(arg[3]);
+        if (!is_const(d,CSTstr))
+          die(n,"Incorrect library name in install");
+        gp->proto.origin=entryname(d);
       }
       d=detag(arg[1]);
       if (tree[d].f!=Fentry && !is_const(d,CSTstr))
@@ -287,16 +288,21 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       tree[d].f=Fconst;
       p=entryname(d);
       gp->proto.code=p+1;
+      funcmode(*gp)=0;
       switch(p[0])
       {
       case 'v':
-	functype(*gp)=Gvoid;
+        functype(*gp)=Gvoid;
         break;
       case 'l':
-	functype(*gp)=Gsmall;
+        functype(*gp)=Gsmall;
         break;
       case 'i':
-	functype(*gp)=Gsmall_int;
+        functype(*gp)=Gsmall_int;
+        break;
+      case 'm':
+        functype(*gp)=Ggen;
+        funcmode(*gp)=(1<<Mcopy);
         break;
       default:
         functype(*gp)=Ggen;
@@ -304,7 +310,6 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       }
       if (tree[arg[1]].f==Ftag)
         functype(*gp)=tree[arg[1]].y;
-      funcmode(*gp)=0;
       for(;*p;p++) if (*p=='p') funcmode(*gp)|=(1<<Mprec);
       gp->spec=GPinstalled;
     }
@@ -318,21 +323,18 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vsup);
     bseq=addseqright(bseq,geninsertvar(bsup,newleaf(vsup)));
     genblock(bseq,-1);
-    savx=s_ctx.n; 
+    savx=s_ctx.n;
     bl1=newblock();
     bl2=newblock();
     if (autogc)
       gpm=gengerepile(bl2);
     genequaltosmall(binf,bsup,&tv);
     pushvar(var,1<<Cuser,tv,-1);
-    lab=genblock_label(seq,-1);
+    lab=makesubblock_label(seq);
     if (lab.ne)
       seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
-    {
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-      gengerepileend(bl2);
-    }
     arg[0]=newnode(Faffect,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
@@ -354,7 +356,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       vfvnext=newnode(Fentry,newentry("fv_next"),-1);
       bseq=addseqright(bseq,geninsertvar(binf,newleaf(vinf)));
       genblock(bseq,-1);
-      sav0=s_ctx.n; 
+      sav0=s_ctx.n;
       bl0=newblock();
       pushvar(vfvdat,(1<<Cuser)|(1<<Cconst),Ggen,-1);
       pushvar(vfvnext,(1<<Cuser)|(1<<Cconst),Gfunc_GG,-1);
@@ -363,7 +365,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
               newleaf(vinf),newleaf(flag)),
               newnode(Frefarg,newnode(Fentry,newentry("fv_data"),-1),-1)),
               newnode(Frefarg,newnode(Fentry,newentry("fv_next"),-1),-1))));
-      savx=s_ctx.n; 
+      savx=s_ctx.n;
       bl1=newblock();
       bl2=newblock();
       if (autogc)
@@ -371,14 +373,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       next = newnode(Faffect,newleaf(var),newnode(Ftag,
           newcall("_call_GG",newnode(Flistarg,newnode(Flistarg,
               newleaf(vfvnext),newleaf(vfvdat)),newleaf(var))),Gvec));
-      lab=genblock_label(seq,-1);
+      lab=makesubblock_label(seq);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
-      {
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
-        gengerepileend(bl2);
-      }
       arg[0]=GNOARG;
       arg[1]=newleaf(var);
       arg[3]=next;
@@ -400,7 +399,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       div=newcall("divisors",arg[0]);
       bseq=addseqright(bseq,geninsertvar(div,vdiv));
       genblock(bseq,-1);
-      savx=s_ctx.n; 
+      savx=s_ctx.n;
       bl1=newblock();
       bl2=newblock();
       if (autogc)
@@ -408,14 +407,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       newdecl(0,Gsmall,-1,&vloop);
       if (tv==Gnotype) tv=Gint;
       pushvar(var,1<<Cuser,tv,-1);
-      lab=genblock_label(seq,-1);
+      lab=makesubblock_label(seq);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
-      {
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
-        gengerepileend(bl2);
-      }
       seq=addseqleft(
           newnode(Faffect,newleaf(var),
             newnode(Ftag,
@@ -466,14 +462,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       pushvar(var,1<<Cuser,tv,-1);
       if (dosign)
         newdecl(0,Gbool,newopcall(OPg,newleaf(vstep),0),&vsign);
-      lab=genblock_label(seq,-1);
+      lab=makesubblock_label(seq);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
-      {
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
-        gengerepileend(bl2);
-      }
       arg[0]=newnode(Faffect,var,vinf);
       if (isvec)
       {
@@ -489,7 +482,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         {
           if ( tree[bstep].x>0 )
             arg[3]=newopcall(OPle,newleaf(var),vsup);
-          else  
+          else
             arg[3]=newopcall(OPge,newleaf(var),vsup);
           nb=4;
         }
@@ -526,9 +519,9 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       newdecl(0,Ggen,-1,&ret);
       aseq=addseqright(aseq,geninsertvar(
             newnode(Fconst,newsmallrealvalue(1),-1),newleaf(ret)));
-    }    
+    }
     genblock(bseq,-1);
-    savx=s_ctx.n; 
+    savx=s_ctx.n;
     bl1=newblock();
     bl2=newblock();
     if (autogc)
@@ -541,22 +534,20 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     if (gp->spec==GPprodeuler)
     {
       seq=geninsertvarop(seq,ret,OPme);
-      genblock_nolabel(seq,-1);
+      makesubblock_nolabel(seq);
+      lab.go=0;
     }
     else
     {
-      lab=genblock_label(seq,-1);
+      lab=makesubblock_label(seq);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     }
     if (autogc)
-    {
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-      gengerepileend(bl2);
-    }
     /*if (vsup>maxprime()) err(primer1);*/
     aseq=addseqleft(newcall("if",newnode(Flistarg,
-              newopcall(OPg,vsup,newcall("_maxprime",-1)), 
+              newopcall(OPg,vsup,newcall("_maxprime",-1)),
               newcall("_err_primes",-1)))
               ,aseq);
     if (tree[binf].f!=Fsmall || tree[binf].x>2)
@@ -615,12 +606,9 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     genequaltosmall(binf,bsup,&tv);
     pushvar(var,1<<Cuser,tv,-1);
     seq=geninsertvarop(seq,ret,(gp->spec==GPsum?OPpe:OPme));
-    genblock_nolabel(seq,-1);
+    makesubblock_nolabel(seq);
     if (autogc)
-    {
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-      gengerepileend(bl2);
-    }
     arg[0]=newnode(Faffect,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
@@ -650,12 +638,9 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (tv==Gnotype) tv=Gint;
       pushvar(var,1<<Cuser,tv,-1);
       seq=geninsertvarop(seq,ret,OPpe);
-      genblock_nolabel(seq,-1);
+      makesubblock_nolabel(seq);
       if (autogc)
-      {
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
-        gengerepileend(bl2);
-      }
       seq=addseqleft(
           newnode(Faffect,newleaf(var),
             newnode(Ftag,
@@ -664,7 +649,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       arg[1]=newopcall(OPle,newleaf(vloop),newcall("length",vdiv));
       arg[2]=seq;
       arg[3]=newopcall(OPpp,newleaf(vloop),-1);
-      aseq=newnode(tree[n].f,newentry("sum"),listtoseq(arg,4));
+      aseq=newnode(tree[n].f,newentry("for"),listtoseq(arg,4));
       makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
       break;
     }
@@ -677,14 +662,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     if(autogc)
       gpm=gengerepile(bl1);
     genblock_nolabel(arg[0],n);
-    lab=genblock_label(seq,-1);
+    lab=makesubblock_label(seq);
     if (lab.ne)
       seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
-    {
-      gengerepileend(bl1);
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    }
     arg[1]=seq;
     aseq=newnode(tree[n].f,tree[n].x,listtoseq(arg,nb));
     if (lab.go)
@@ -716,7 +698,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=newcoeff(newleaf(ret),newleaf(var),-1);
     arg[2]=geninsertvar(seq,arg[2]);
-    genblock_nolabel(arg[2],-1);
+    makesubblock_nolabel(arg[2]);
     arg[3]=newopcall(OPpp,newleaf(var),-1);
     aseq=addseqright(aseq,newcall("for",listtoseq(arg,4)));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
@@ -761,7 +743,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       arg[3]=newopcall(OPpp,newleaf(var2),-1);
       arg[2]=newcoeff(newleaf(ret),newleaf(var2),newleaf(var1));
       arg[2]=geninsertvar(seq,arg[2]);
-      genblock_nolabel(arg[2],-1);
+      makesubblock_nolabel(arg[2]);
       arg[2]=addseqright(aseq,newcall("for",listtoseq(arg,4)));
       aseq=newgetg(ret,newleaf(vsup1),"t_MAT");
       arg[0]=newnode(Faffect,var1,newsmall(1));
@@ -774,16 +756,15 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
   case GPglobal:
   case GPlocal:
     {
-      int narg;
       int aseq=GNIL;
       int flag=1<<Cuser;
       int mint=(tree[p].f==Ftag)?tree[p].y:-1;
       if (gp->spec==GPglobal)
         flag|=1<<Cglobal;
       savx=s_ctx.n;
-      narg=genblockdeclaration(y,n,flag,mint);
-      aseq=genblockinit(aseq,savx,narg,gp->spec==GPglobal);
+      (void)genblockdeclaration(y,n,flag,mint,&aseq);
       makeblock(newblock(),n,aseq,-1,s_ctx.n);
+      if (tree[p].f==Ftag) tree[p]=tree[n];
     }
     break;
   case GPreturn:
@@ -796,7 +777,6 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     if (tree[y].f==Fentry)
     {
       var=y;
-      vret=getvar(var);
       bseq=-1;
       y=newleaf(var);
     }
@@ -804,14 +784,13 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     {
       /*Use the block value variable as the return variable*/
       var=block[tree[y].x].ret;
-      vret=var<0?-1:getvar(var);
       bseq=y;
       y=newleaf(var);
     }
     else if (tree[y].f!=Fsmall)
     {
       /*Use the returned variable as the return variable*/
-      vret=newdecl((1<<Cauto),Gempty,-1,&var);
+      newdecl((1<<Cauto),Gempty,-1,&var);
       bseq=geninsertvar(y,var);
       y=newleaf(var);
     }
@@ -819,15 +798,12 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     bl2=newblock();
     savx=s_ctx.n;
     block[bl2].gc|=(1<<GCneeded)|(1<<GCglobal)|(1<<GCreturn);
-    block[bl2].egc=vret;
+    block[bl2].egc=var;
     arg[0]=newnode(Fentry,newentry("ltop"),-1);
     arg[1]=GNOARG;
-    arg[2]=(FC_gerepileall<0)?newnode(Fentry,newentry("gptr"),-1):GNOARG;
-    arg[3]=newsmall(bl2);
-    gpm=listtoseq(arg,4);
+    arg[2]=newsmall(bl2);
+    gpm=listtoseq(arg,3);
     aseq=addseqright(aseq,newcall("_gerepilemany",gpm));
-    if (FC_gerepileall<0)
-      pushvar(newleaf(arg[2]),1<<Cuser,Ggptr,newcall("_gerepilelist",arg[3]));
     aseq=addseqright(aseq,newnode(tree[n].f,tree[n].x,y));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     break;
@@ -850,7 +826,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       l=s_label.n;
       k=nb;
       while(k--)
-        if (!label[--l].ok)
+        if (!l || !label[--l].ok)
           die(n,"%s not allowed here",gp->gpname);
       if (nb>1)
       {
@@ -897,18 +873,18 @@ int gentypefuncspec(int n, gpfunc *gp)
     nb=genlistargs(n,arg,2,3);
     gentype(arg[0]);
     gentype(arg[1]);
-    tree[n].m=(tree[arg[0]].m|(tree[arg[1]].m&~(1<<Mterm)))&MODHERIT;
+    tree[n].m=(1<<Melse)|((tree[arg[0]].m|(tree[arg[1]].m&~(1<<Mterm)))&MODHERIT);
     if (nb==3)
     {
       gentype(arg[2]);
-      /*We must not set Gterm if we are not sure.
-        In this case we must set Gsidef instead.*/
-      tree[n].m|=(1<<Melse)|((tree[arg[2]].m&(~(1<<Mterm)))&MODHERIT);
+      /*We must not set Mterm if we are not sure.
+        In this case we must set Msidef instead.*/
+      tree[n].m|=((tree[arg[2]].m&(~(1<<Mterm)))&MODHERIT);
       if ((tree[arg[1]].m&(1<<Mterm)) && (tree[arg[2]].m&(1<<Mterm)))
-	tree[n].m|=(1<<Mterm);
+        tree[n].m|=(1<<Mterm);
       else
-	if ((tree[arg[1]].m&(1<<Mterm)) || (tree[arg[2]].m&(1<<Mterm)))
-	  tree[n].m|=(1<<Msidef);
+        if ((tree[arg[1]].m&(1<<Mterm)) || (tree[arg[2]].m&(1<<Mterm)))
+          tree[n].m|=(1<<Msidef);
       return typemax[tree[arg[1]].t][tree[arg[2]].t];
     }
     if ((tree[arg[1]].m&(1<<Mterm)))
@@ -937,7 +913,13 @@ int gentypefuncspec(int n, gpfunc *gp)
     if(gp->spec==GPwhile)
       tree[n].m=(tree[arg[0]].m|(tree[arg[1]].m&~(1<<Mterm)))&MODHERIT;
     else
-      tree[n].m=(tree[arg[0]].m|tree[arg[1]].m)&MODHERIT;
+    {
+      int a = arg[1];
+      tree[n].m=((tree[arg[0]].m|tree[arg[1]].m)&MODHERIT);
+      tree[a].m|=(1<<Muntil);
+      if (tree[a].f==Fblock)
+        tree[tree[a].y].m|=(1<<Muntil);
+    }
     return Gvoid;
   case GPreturn:
     {
@@ -968,15 +950,21 @@ int gentypefuncspec(int n, gpfunc *gp)
     tree[n].m=0;
     return Gvoid;
   case GPmakevec:
-  case GPmakemat:
     gentype(y);
     tree[n].m=(tree[y].m&MODHERIT)|(1<<Msidef);
     return Gvec;
-  case GPgplist:
-    tree[n].m=0;
-    return Ggptr;
+  case GPclosure:
+    nb=genlistargs(n,arg,1,STACKSZ);
+    gentypemode(nb,arg,&tree[n].m);
+    if ((tree[n].m&(1<<Mprec))==0)
+    {
+      int nf = findfunction(entryname(arg[0]));
+      if (nf>=0 && lfunc[nf].spec==GPuser && lfunc[nf].user->wrapper>=0 && (funcmode(lfunc[nf])&(1<<Mprec)))
+        tree[n].m|=(1<<Mprec);
+    }
+    return Gclosure;
   default:
-    if (gp->spec>0) 
+    if (gp->spec>0)
       die(n,"gentypefuncspec : func spec not implemented");
     else
       die(n,"Internal error : gentypefuncspec called with no spec");
@@ -997,14 +985,14 @@ void gencopyarg(FILE *fout)
       ctxvar *v=fc->c+i;
       long j;
       for(j=0;j<fc->v.n;j++)
-	if (fc->var[j].f==AFaffectcompo && fc->var[j].idx==i+fc->savb-fc->s.n )
-	{
-	  genindent(fout);
-	  fprintf(fout,"%s = ",v->cvar);
+        if (fc->var[j].f==AFaffectcompo && fc->var[j].idx==i+fc->savb-fc->s.n )
+        {
+          genindent(fout);
+          fprintf(fout,"%s = ",v->cvar);
           genfuncbydesc1(fout,v->node,FC_copy,n);
-          gensemicomma(fout,v->node);
-	  break;
-	}
+          gensemicolon(fout,v->node);
+          break;
+        }
     }
 }
 
@@ -1034,7 +1022,6 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
   int nb;
   int y=tree[n].y;
   int i;
-  int ncol,nlin;
   context *bl;
   switch(gp->spec)
   {
@@ -1047,186 +1034,183 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       break;
     case GPif:
       nb=genlistargs(n,arg,2,3);
-#if 0    
+#if 0
       /*generate nice `? :' */
-    if (t!=Gvoid && tree[arg[1]].f!=Fseq && (nb==2 || tree[arg[2]].f!=Fseq) )
-    {
-      fprintf(fout,"((");
-      genbool(fout,arg[0],tree[arg[0]].t);
-      fprintf(fout,")?");
-      gencast(fout,arg[3],t);
-      fprintf(fout,":");
-      if (nb==5)
-	gencast(fout,arg[4],t);
-      else/*the type of left and right must be equal.*/
-	gensmallval(fout,0,t);
-      fprintf(fout,")");
-    }
-#endif
-    genindent(fout);
-    fprintf(fout,"if (");
-    gencast(fout,arg[0],Gbool);
-    fprintf(fout,")\n");
-    if (nb==2)
-      genbrace(fout,arg[1]);
-    else
-    {
-      /*Be paranoid with dangling 'else' */
-      int brace=(tree[arg[1]].f==Fseq || tree[arg[1]].f==Fblock);
-      if (brace)
+      if (t!=Gvoid && tree[arg[1]].f!=Fseq && (nb==2 || tree[arg[2]].f!=Fseq) )
       {
+        fprintf(fout,"((");
+        genbool(fout,arg[0],tree[arg[0]].t);
+        fprintf(fout,")?");
+        gencast(fout,arg[3],t);
+        fprintf(fout,":");
+        if (nb==5)
+          gencast(fout,arg[4],t);
+        else/*the type of left and right must be equal.*/
+          gensmallval(fout,0,t);
+        fprintf(fout,")");
+      }
+#endif
+      genindent(fout);
+      fprintf(fout,"if (");
+      gencast(fout,arg[0],Gbool);
+      fprintf(fout,")\n");
+      if (nb==2)
+        genbrace(fout,arg[1]);
+      else
+      {
+        genbrace(fout,arg[1]);
+        genindent(fout);
+        fprintf(fout,"else\n");
+        genbrace(fout,arg[2]);
+      }
+      break;
+    case GPwhile:
+      nb=genlistargs(n,arg,2,2);
+      genindent(fout);
+      if (tree[arg[0]].f!=Fseq && tree[arg[0]].f!=Fblock)
+      {
+        fprintf(fout,"while (");
+        gencast(fout,arg[0],Gbool);
+        fprintf(fout,")\n");
+        genbrace(fout,arg[1]);
+      }
+      else
+      {
+        int x=tree[arg[0]].x;
+        int y=tree[arg[0]].y;
+        if (tree[arg[0]].f==Fblock)
+        {
+          y=block[x].ret;
+          x=tree[arg[0]].y;
+        }
+        fprintf(fout,"for(;;)\n");
         genindent(fout);
         fprintf(fout,"{\n");
-      }
-      indent++;
-      genindentseq(fout,arg[1]);
-      gencodeg(fout,arg[1]);
-      gensemicomma(fout,arg[1]);
-      indent--;
-      if (brace)
-      {
+        indent++;
+        genindentseq(fout,x);
+        gencodeg(fout,x);
+        gensemicolon(fout,x);
+        if (tree[y].f==Fblock)
+        {
+          gencodeg(fout,y);
+          y=block[tree[y].x].ret;
+        }
+        genindent(fout);
+        fprintf(fout,"if (");
+        gencast(fout,y,Gnegbool);
+        fprintf(fout,")\n");
+        indent++;
+        genindent(fout);
+        fprintf(fout,"break;\n");
+        indent--;
+        genindentseq(fout,arg[1]);
+        gencodeg(fout,arg[1]);
+        gensemicolon(fout,arg[1]);
+        indent--;
         genindent(fout);
         fprintf(fout,"}\n");
       }
+      break;
+    case GPuntil:
+      nb=genlistargs(n,arg,2,2);
       genindent(fout);
-      fprintf(fout,"else\n");
+      if ( tree[arg[0]].f!=Fseq && tree[arg[0]].f!=Fblock)
+      {
+        fprintf(fout,"do\n");
+        genbrace(fout,arg[1]);
+        fprintf(fout," while(");/*KB style*/
+        gencast(fout,arg[0],Gnegbool);
+        fprintf(fout,");\n");
+      }
+      else
+      {
+        int x=tree[arg[0]].x;
+        int y=tree[arg[0]].y;
+        if (tree[arg[0]].f==Fblock)
+        {
+          y=block[x].ret;
+          x=tree[arg[0]].y;
+        }
+        fprintf(fout,"for(;;)\n");
+        genindent(fout);
+        fprintf(fout,"{\n");
+        indent++;
+        genindentseq(fout,arg[1]);
+        gencodeg(fout,arg[1]);
+        gensemicolon(fout,arg[1]);
+        genindentseq(fout,x);
+        gencodeg(fout,x);
+        gensemicolon(fout,x);
+        genindent(fout);
+        if (tree[y].f==Fblock)
+        {
+          gencodeg(fout,y);
+          y=block[tree[y].x].ret;
+        }
+        fprintf(fout,"if (");
+        gencast(fout,y,Gbool);
+        fprintf(fout,")\n");
+        indent++;
+        genindent(fout);
+        fprintf(fout,"break;\n");
+        indent--;
+        indent--;
+        genindent(fout);
+        fprintf(fout,"}\n");
+      }
+      break;
+    case GPfor:
+      nb=genlistargs(n,arg,4,4);
+      genindent(fout);
+      fprintf(fout,"for (");
+      if (arg[0]!=GNOARG)
+        gencode(fout,arg[0]);
+      else
+        fprintf(fout,"  ");
+      fprintf(fout,"; ");
+      if (arg[1]!=GNOARG)
+        gencode(fout,arg[1]);
+      fprintf(fout,"; ");
+      if (arg[3]!=GNOARG)
+        gencode(fout,arg[3]);
+      fprintf(fout,")\n");
       genbrace(fout,arg[2]);
-    }
-    break;
-  case GPwhile:
-    nb=genlistargs(n,arg,2,2);
-    genindent(fout);
-    if (tree[arg[0]].f!=Fseq)
-    {
-      fprintf(fout,"while (");
-      gencast(fout,arg[0],Gbool);
-      fprintf(fout,")\n");      
-      genbrace(fout,arg[1]);
-    }
-    else
-    {
-      int x=tree[arg[0]].x;
-      int y=tree[arg[0]].y;
-      fprintf(fout,"for(;;)\n");
+      break;
+    case GPdoloop:
       genindent(fout);
-      fprintf(fout,"{\n");
-      indent++;
-      genindentseq(fout,x);
-      gencodeg(fout,x);
-      gensemicomma(fout,x);
+      fprintf(fout,"for (;;)\n");
+      genbrace(fout,y);
+      break;
+    case GPforstep:
+      nb=genlistargs(n,arg,4,6);
       genindent(fout);
-      fprintf(fout,"if (");
-      gencast(fout,y,Gnegbool);
-      fprintf(fout,")\n");
-      indent++;
-      genindent(fout);
-      fprintf(fout,"break;\n");
-      indent--;
-      genindentseq(fout,arg[1]);
-      gencodeg(fout,arg[1]);
-      gensemicomma(fout,arg[1]);
-      indent--;
-      genindent(fout);
-      fprintf(fout,"}\n");
-    }
-    break;
-  case GPuntil:
-    nb=genlistargs(n,arg,2,2);
-    genindent(fout);
-    if ( tree[arg[0]].f!=Fseq )
-    {
-      fprintf(fout,"do\n");
-      genindent(fout);
-      fprintf(fout,"{\n");
-      indent++;
-      genindentseq(fout,arg[1]);
-      gencodeg(fout,arg[1]);
-      gensemicomma(fout,arg[1]);
-      indent--;
-      genindent(fout);
-      fprintf(fout,"} while(");/*KB style*/
-      gencast(fout,arg[0],Gnegbool);
-      fprintf(fout,");\n");
-    }
-    else
-    {
-      int x=tree[arg[0]].x;
-      int y=tree[arg[0]].y;
-      fprintf(fout,"for(;;)\n");
-      genindent(fout);
-      fprintf(fout,"{\n");
-      indent++;
-      genindentseq(fout,arg[1]);
-      gencodeg(fout,arg[1]);
-      gensemicomma(fout,arg[1]);
-      genindentseq(fout,x);
-      gencodeg(fout,x);
-      gensemicomma(fout,x);
-      genindent(fout);
-      fprintf(fout,"if (");
-      gencast(fout,y,Gbool);
-      fprintf(fout,")\n");
-      indent++;
-      genindent(fout);
-      fprintf(fout,"break;\n");
-      indent--;
-      indent--;
-      genindent(fout);
-      fprintf(fout,"}\n");
-    }
-    break;
-  case GPfor:
-    nb=genlistargs(n,arg,4,4);
-    genindent(fout);
-    fprintf(fout,"for (");
-    if (arg[0]!=GNOARG)
+      fprintf(fout,"for (");
       gencode(fout,arg[0]);
-    else
-      fprintf(fout,"  ");
-    fprintf(fout,"; ");
-    if (arg[1]!=GNOARG)
-      gencode(fout,arg[1]);
-    fprintf(fout,"; ");
-    if (arg[3]!=GNOARG)
-      gencode(fout,arg[3]);
-    fprintf(fout,")\n");
-    genbrace(fout,arg[2]);
-    break;
-  case GPdoloop:
-    genindent(fout);
-    fprintf(fout,"for (;;)\n");
-    genbrace(fout,y);
-    break;
-  case GPforstep:
-    nb=genlistargs(n,arg,4,6);
-    genindent(fout);
-    fprintf(fout,"for (");
-    gencode(fout,arg[0]);
-    fprintf(fout,"; ");
-    if ( nb==4 )
-      gencode(fout,arg[3]);
-    else
-    {
-      gencode(fout,arg[3]);
-      fprintf(fout,"?");
-      gencode(fout,arg[4]);
-      fprintf(fout,":");
-      gencode(fout,arg[5]);
-    }
-    fprintf(fout,"; ");
-    gencode(fout,arg[2]);
-    fprintf(fout,")\n");
-    genbrace(fout,arg[1]);
-    break;
+      fprintf(fout,"; ");
+      if ( nb==4 )
+        gencode(fout,arg[3]);
+      else
+      {
+        gencode(fout,arg[3]);
+        fprintf(fout,"?");
+        gencode(fout,arg[4]);
+        fprintf(fout,":");
+        gencode(fout,arg[5]);
+      }
+      fprintf(fout,"; ");
+      gencode(fout,arg[2]);
+      fprintf(fout,")\n");
+      genbrace(fout,arg[1]);
+      break;
 #if 0
-    forstep(x=binf,bsup,step,SEQ)
-    {
-      int idx;
+      forstep(x=binf,bsup,step,SEQ)
+      {
+        int idx;
 
       for( idx=0,x=binf; x<=bsup; i=i==lg(step)?1:i+1,x+=step[i])
       {
       }
-    }    
+    }
 #endif
   case GPforstepvec:
     nb=genlistargs(n,arg,6,6);
@@ -1254,51 +1238,17 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     nb=genlistargs(n,arg,1,STACKSZ);
     genindent(fout);
     gencode(fout,arg[0]);
-    gensemicomma(fout,arg[0]);
+    gensemicolon(fout,arg[0]);
     for(i=1;i<nb;i++)
     {
       genindentseq(fout,arg[i]);
       gencodeg(fout,arg[i]);
-      gensemicomma(fout,arg[i]);
+      gensemicolon(fout,arg[i]);
     }
     break;
-  case GPmakemat:
-    nb=genlistargs(n,arg,3,STACKSZ);
-    nlin=tree[arg[1]].x;
-    ncol=tree[arg[2]].x;
-    genindent(fout);
-    gencode(fout,arg[0]);
-    gensemicomma(fout,arg[0]);
-    for(i=1;i<=ncol;i++)
-    {
-      int j;
-      genindent(fout);
-      gencode(fout,arg[i+2]);
-      gensemicomma(fout,arg[i+2]);
-      indent++;
-      for(j=1;j<=nlin;j++)
-      {
-        int k=(j-1)*ncol+i+2+ncol;
-        genindentseq(fout,arg[k]);
-        gencodeg(fout,arg[k]);
-        gensemicomma(fout,arg[k]);
-      }
-      indent--;
-    }
-    break;
-  case GPgplist:
-    bl=block+tree[y].x;
-    fprintf(fout,"{ ");
-    for(i=0;i<bl->g.n;i++)
-    {
-      if (i) fprintf(fout,", ");
-      fprintf(fout,"&%s",ctxstack[bl->gcvar[i]].cvar);
-    }
-    fprintf(fout," }");
-    break;
   case GPgpmany:
-    nb=genlistargs(n,arg,4,STACKSZ);
-    bl=block+tree[arg[3]].x;
+    nb=genlistargs(n,arg,3,STACKSZ);
+    bl=block+tree[arg[2]].x;
     if (!(bl->gc&(1<<GCneeded)))
       break;
     genindent(fout);
@@ -1311,20 +1261,20 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     }
     else if (bl->gc&(1<<GCupto))
     {
-      int var=ctxstack[bl->gcvar[0]].node;
+      int var=bl->gcvar[0];
       int parg[2];
       gencode(fout,var);
       fprintf(fout," = ");
       parg[0]=arg[0]; parg[1]=var;
       genfuncbydesc(fout,2,parg,FC_gerepileupto,n);
-      fprintf(fout,";\n"); 
+      fprintf(fout,";\n");
     }
     else
     {
-      /*FIXME:Putting an if statement in the wilderness is not secure*/ 
+      int parg[STACKSZ+1];
+      /*FIXME:Putting an if statement in the wilderness is not secure*/
       if (arg[1]!=GNOARG)
       {
-        int parg[2];
         parg[0]=arg[1]; parg[1]=arg[0];
         fprintf(fout,"if (");
         genfuncbydesc(fout,2,parg,FC_low_stack_lim,n);
@@ -1333,23 +1283,11 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
         genindent(fout);
         indent--;
       }
-      if (arg[2]!=GNOARG)
-      {
-        fprintf(fout,"gerepilemany(");
-        gencode(fout,arg[0]);
-        fprintf(fout,", ");
-        gencode(fout,arg[2]);
-        fprintf(fout,", %d);\n",bl->g.n);
-      }
-      else
-      {
-        int parg[STACKSZ+1];
-        parg[0]=arg[0];
-        for(i=0;i<bl->g.n;i++)
-          parg[i+1]=ctxstack[bl->gcvar[i]].node;
-        genfuncbydesc(fout,bl->g.n+1,parg,FC_gerepileall,n);
-        fprintf(fout,";\n");
-      }
+      parg[0]=arg[0];
+      for(i=0;i<bl->g.n;i++)
+        parg[i+1]=bl->gcvar[i];
+      genfuncbydesc(fout,bl->g.n+1,parg,FC_gerepileall,n);
+      fprintf(fout,";\n");
     }
     break;
   case GPreturn:
@@ -1373,8 +1311,15 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
   case GPaddhelp:
   case GPinstall:
     break;
+  case GPclosure:
+    nb=genlistargs(n,arg,1,STACKSZ);
+    if (FC_strtoclosure >=0)
+      genfuncbydesc(fout,nb,arg,FC_strtoclosure,n);
+    else
+      die(n,"closure not available in this version");
+    break;
   default:
-    if (gp->spec>0) 
+    if (gp->spec>0)
       die(n,"genentryspec: func spec not implemented");
     else
       die(n,"Internal error: genentryspec called with no spec");
