@@ -32,7 +32,7 @@ int newdecl(int flag, int t, int initval, int *v)
 }
 int newcall(const char *s, int y)
 {
-  return newnode(Fentryfunc,newentry(s),y);
+  return newnode(Ffunction,newentry(s),y);
 }
 
 int listtoseq(int *stack, int nb)
@@ -46,11 +46,11 @@ int listtoseq(int *stack, int nb)
 int newseq(int x, int y)
 {
   int seq;
-  if (tree[x].f==Fgnil || tree[y].f==Fgnil)
+  if (tree[x].f==Fnoarg || tree[y].f==Fnoarg)
   {
-    if (tree[x].f==Fgnil && tree[y].f==Fgnil)
+    if (tree[x].f==Fnoarg && tree[y].f==Fnoarg)
       return GNIL;
-    else if (tree[x].f==Fgnil)
+    else if (tree[x].f==Fnoarg)
       return y;
     else
       return x;
@@ -59,6 +59,19 @@ int newseq(int x, int y)
   return seq;
 }
 
+int addlistleft(int n, int l)
+{
+  int i,b;
+  if (l==GNOARG)
+    return n;
+  if (tree[l].f!=Flistarg)
+    return newnode(Flistarg,n,l);
+  for(i=l;tree[tree[i].x].f==Flistarg;i=tree[i].x);
+  b=newnode(Flistarg,n,tree[i].x);
+  tree[i].x=b;
+  return l;
+}
+
 int addseqleft(int n, int seq)
 {
   int i;
@@ -73,7 +86,7 @@ int addseqleft(int n, int seq)
   return seq;
 }
 int addseqright(int seq, int n)
-{ 
+{
   if(seq==-1)
     return n;
   return newnode(Fseq,seq,n);
@@ -85,13 +98,13 @@ static int insertreturn(int y)
     return y;
   return newcall("return",y);
 }
-  
+
 int geninsertreturn(int seq)
 {
   int bseq;
-  if (seq<0) return seq;
+  if (seq==GNOARG) return seq;
   if (tree[seq].f==Fseq)
-  {  
+  {
     bseq=insertreturn(tree[seq].y);
     bseq=newnode(Fseq,tree[seq].x,bseq);
   }
@@ -108,7 +121,7 @@ int geninsertvar(int seq, int ret)
   if (ret>=0)
   {
     if (tree[seq].f==Fseq)
-    {  
+    {
       bseq=newnode(Faffect,ret,tree[seq].y);
       bseq=newnode(Fseq,tree[seq].x,bseq);
     }
@@ -127,7 +140,7 @@ int geninsertvarop(int seq, int ret, OPerator op)
   if (ret>=0)
   {
     if (tree[seq].f==Fseq)
-    {  
+    {
       bseq=newopcall(op,ret,tree[seq].y);
       bseq=newnode(Fseq,tree[seq].x,bseq);
     }
@@ -144,7 +157,7 @@ int newcoeff(int n, int x, int y)
   int arg[3];
   int nb=y<0?2:3;
   arg[0]=n; arg[1]=x; arg[2]=y;
-  return newnode(Fentryfunc,OPcoeff,listtoseq(arg,nb));
+  return newnode(Ffunction,OPcoeff,listtoseq(arg,nb));
 }
 
 void makeblock(int bl, int n, int aseq, int ret, int savx)
@@ -189,22 +202,27 @@ void genblock_nolabel(int n, int p)
   stack_pop_safe(&s_label,l);
 }
 
-label_t genblock_label(int n, int p)
+Clabel_t genblock_label(int n, int p)
 {
-  label_t t;
+  Clabel_t t;
   int l=newlabel(1);
   genblock(n,p);
   t=label[l];
   stack_pop_safe(&s_label,l);
   return t;
 }
+static void affectval(int vn, int val, int *seq)
+{
+  ctxvar *v = ctxstack+vn;
+  int valn=newnode(Faffect,newleaf(v->node),val);
+  *seq=addseqright(*seq,valn);
+}
 
-int genblockdeclaration(int args, int n, int flag, int type)
+int genblockdeclaration(int args, int n, int flag, int type, int *seq)
 {
   int stack[STACKSZ];
   int i;
   enum {local,global,function} decl;
-  int savx=s_ctx.n;
   int mint=(type>=0)?type:Ggen;
   int nb=listtostack(args,Flistarg,stack,STACKSZ,"function declaration",n);
   if (nb==1 && stack[0]==GNOARG)
@@ -223,53 +241,71 @@ int genblockdeclaration(int args, int n, int flag, int type)
         die(stack[i],"Incorrect declaration");
       if (decl!=function && autogc && ctype[tv]==Vgen)
         /*Make sure GEN objects are gerepilable.*/
-        val=newsmall(0);
-      pushvar(var,flag,tv,val);
+        val = newnode(Ftag, newnode(Ftag, newsmall(0), Ggen), tv);
+      if (decl==global)
+        fillvar(var,flag,tv,val);
+      else
+        pushvar(var,flag,tv,val);
       break;
     case Fentry:
       switch(decl)
       {
       case local:
         if (type<0 || autogc)
-        /*Make sure (implicitly GEN) objects are gerepilable.*/
+          /*Make sure (implicitly GEN) objects are gerepilable.*/
           val=newsmall(0);
+        pushvar(stack[i],flag,mint,val);
         break;
       case global:
         if (type<0)
+        {
           val=newcall("_const_quote",newstringnode(entryname(stack[i]),-1));
+          affectval(fillvar(stack[i],flag,mint,-1),val,seq);
+        }
         else if (autogc)
-        /*Make sure (implicitly GEN) objects are gerepilable.*/
-          val=newsmall(0);
+        {
+          /*Make sure (implicitly GEN) objects are gerepilable.*/
+          fillvar(stack[i],flag,mint, newsmall(0));
+        }
         break;
       case function:
         if (!optstrict)
           val=newsmall(0);
+        pushvar(stack[i],flag,mint,val);
         break;
       }
-      pushvar(stack[i],flag,mint,val);
       break;
     case Faffect:
+      genequal(stack[i],"declaration",&var,&val,&tv);
+      if (tv==Gnotype)
+        tv=mint;
+      switch (decl)
       {
-        ctxvar *v;
-        genequal(stack[i],"declaration",&var,&val,&tv);
-        if (tv==Gnotype)
-          tv=mint;
-        v=ctxstack+pushvar(var,flag,tv,val);
-        if (decl==function && descfindrules1(v->node,FC_default_check))
-          v->flag|=(1<<Cdefmarker);
+      case global:
+        affectval(fillvar(var,flag,tv,-1),val,seq);
+        break;
+      case local:
+        if (tree[val].f==Fsmall)
+          pushvar(var,flag,tv,val);
+        else
+          affectval(pushvar(var,flag,tv,-1),val,seq);
+        break;
+      case function:
+        {
+          int vn = pushvar(var,flag,tv,val);/* can change ctxstack */
+          ctxvar *v = ctxstack+vn;
+          if (decl==function && descfindrules1(v->node,FC_default_check))
+            v->flag|=(1<<Cdefmarker);
+        }
+        break;
       }
+      genblock(val,n);
       break;
     default:
       die(n,"Incorrect node %s in function declaration",
           funcname(tree[stack[i]].f));
     }
   }
-  for (i=0; i<nb; i++)
-  {
-    ctxvar *v=ctxstack+savx+i;
-    if (v->initval>=0)
-      genblock(v->initval,n);
-  }
   return nb;
 }
 
@@ -277,7 +313,7 @@ int genblockcheckargs(int seq, userfunc *ufunc)
 {
   int i;
   gpfunc *gp=lfunc+currfunc;
-  context *fc=block+ufunc->bl; 
+  context *fc=block+ufunc->bl;
   seq=addseqleft(newcall("_copyarg",-1),seq);
   for(i=ufunc->narg-1;i>=0;i--)
   {
@@ -288,9 +324,12 @@ int genblockcheckargs(int seq, userfunc *ufunc)
     {
       check=newcall("_badtype",newleaf(v->node));
       if (rule->type!=Gvoid)
-        check=newcall("if",
-            newnode(Flistarg,check,
-                     newcall("_err_type",newstringnode(gp->gpname,-1))));
+      {
+        int name=newstringnode(gp->gpname,-1);
+        int args = descfindrules1(name,findfuncdesc("_err_type")) ? name :
+                     newnode(Flistarg,name,v->node);
+        check=newcall("if", newnode(Flistarg,check, newcall("_err_type", args)));
+      }
     }
     if (v->initval>=0 && (v->flag&(1<<Cdefmarker)))
     {
@@ -313,14 +352,13 @@ is */
 {
   ulong ltop=avma;
   GEN p1;
-  GEN *gptr[]={&p1,&globalsvar};
   {
     GEN args;
     GEN localvars;
     code;
     p1=...;
   }
-  gerepilemany(ltop,gptr,N);
+  gerepileall(ltop,2,&p1,&globalsvar);
   return p1;
 }
 #endif
@@ -344,7 +382,12 @@ void genblockdeffunc(int n)
   if (currfunc<0)
     die(n,"internal error: unknown function %s in genblockdeffunc",name);
   if (lfunc[currfunc].spec!=GPuser)
-    die(n,"internal error: not a user function %s in genblockdeffunc",name);
+  {
+    if (lfunc[currfunc].spec==GPinstalled)
+      die(n,"Cannot redefine install'ed function %s",name);
+    else
+      die(n,"internal error: not a user function %s in genblockdeffunc",name);
+  }
   ufunc=lfunc[currfunc].user;
   /*create external block*/
   bl1=newblock();
@@ -355,18 +398,18 @@ void genblockdeffunc(int n)
   {
     /*declare ltop*/
     gltop=newnode(Fentry,newentry("ltop"),-1);
-    pushvar(gltop,1<<Cuser,Gpari_sp,newcall("_avma",-1));
+    pushvar(gltop,0,Gpari_sp,newcall("_avma",-1));
   }
   /*declare function arguments*/
   ufunc->sarg=s_ctx.n-sav1;
-  ufunc->narg=genblockdeclaration(args,n,(1<<Cuser)|(1<<Carg)|(1<<Ccompo),-1);
+  ufunc->narg=genblockdeclaration(args,n,(1<<Cuser)|(1<<Carg)|(1<<Ccompo),-1,NULL);
   /*Add return at the end*/
   seq=geninsertreturn(seq);
   /*generate block*/
   genblock(seq,-1);
-  if (n!=initnode)
+  if (tree[funcid].x!=initnode)
   {
-    /*FIXME: For now, clear Cglobal flag to avoid initialization 
+    /*FIXME: For now, clear Cglobal flag to avoid initialization
       problem with implicit fetch_user_var().*/
     int i;
     for (i=sav1;i<s_ctx.n;i++)
@@ -382,7 +425,7 @@ void genblockdeffunc(int n)
   tree[n].y=nodebl1;
   /*restore globalvar*/
   currfunc=savcf;
-  savnpv=newprivvar;
+  newprivvar=savnpv;
 }
 
 int newvectoridx(int var, int ind)
@@ -496,7 +539,7 @@ void genblockmatrix(int n)
   int line[STACKSZ];
   int arg[STACKSZ];
   int x=tree[n].x,xx;
-  int i,j,k,nb=3,nbline,nbcol=-1,ret;
+  int i,j,k,nb=1,nbline,nbcol=-1,ret;
   int aseq;
   nbline=listtostack(x,Fmatrixlines,line,STACKSZ,"[...;...;...]",n);
   for(xx=line[0],nbcol=1;tree[xx].f==Fmatrixelts;xx=tree[xx].x,nbcol++);
@@ -510,12 +553,10 @@ void genblockmatrix(int n)
   }
   newdecl(0,Gvec,-1,&ret);
   arg[0]=newgetg(ret,newsmall(nbcol),"t_MAT");
-  arg[1]=newsmall(nbline);
-  arg[2]=newsmall(nbcol);
-  for(j=1;j<=nbcol;j++)
-    arg[j+2]=newgetg(newcoeff(newleaf(ret),newsmall(j),-1),
+  for(i=1;i<=nbcol;i++)
+    arg[i]=newgetg(newcoeff(newleaf(ret),newsmall(i),-1),
                      newsmall(nbline),"t_COL");
-  for(j=1,k=1,i=3+nbcol;i<nb;i++)
+  for(j=1,k=1;i<nb;i++)
   {
     arg[i]=geninsertvar(arg[i],newmatidx(newleaf(ret),k,j));
     if (j==nbcol)
@@ -528,14 +569,14 @@ void genblockmatrix(int n)
   }
   for(i=0;i<nb;i++)
     genblock(arg[i],-1);
-  aseq=newcall("_makemat",listtoseq(arg,nb));
+  aseq=newcall("_makevec",listtoseq(arg,nb));
   makeblock(newblock(),n,aseq,ret,s_ctx.n);
 }
 
 void genblockprotoraw(int n, gpfunc *gp)
 {
   int arg[STACKSZ];
-  const char *p=gp->proto.code, *q=gp->proto.code;
+  const char *p=gp->proto.code;
   char c;
   PPproto mod;
   int nb=genlistargs(n,arg,0,STACKSZ);
@@ -544,9 +585,11 @@ void genblockprotoraw(int n, gpfunc *gp)
   {
     switch(mod)
     {
-    case PPstd:
     case PPdefault:
     case PPdefaultmulti:
+      if (i>=nb || tree[arg[i]].f==Fnoarg)
+        break;
+    case PPstd: /*Fall through */
       if (c=='r' && i<nb)
       {
         if (tree[arg[i]].f==Fentry)
@@ -556,18 +599,19 @@ void genblockprotoraw(int n, gpfunc *gp)
       }
       break;
     default: /* Skip */
+      i--;
       break;
     }
-    q=p;
+    i++;
   }
 }
 /*
   n: node
   p: caller node
-  
+
   Rule of the game:
   genblock must never be used twice on the same node.
-  Using p=-1 is allowed but this assume the value of n is to be discarded.
+  Using p=-1 is allowed but this assumes the value of n is to be discarded.
 
   genblock must not change p.
 */
@@ -581,6 +625,30 @@ int newgetgvec(const char *t)
                 "vec", -1);
 }
 
+int genclosedlist(int nf, int n)
+{
+  int i;
+  gpfunc *gp=lfunc+nf;
+  context *fc=block+gp->user->bl;
+  for (i=0; i<fc->s.n; i++)
+  {
+    ctxvar *v=fc->c+i;
+    if (v->flag&(1<<Cclosed))
+      n = newnode(Flistarg,n,newleaf(v->node));
+  }
+  return n;
+}
+
+
+static int has_mand_arg(const char *code)
+{
+  PPproto mod;
+  char c;
+  while ((mod=parseproto(&code,&c))!=PPend)
+    if(mod==PPstd) return 1;
+  return 0;
+}
+
 void genblock(int n, int p)
 {
   int x,y;
@@ -618,6 +686,7 @@ void genblock(int n, int p)
     if (tree[x].f==Ffacteurmat)
     {
       int l=getlvalue(x);
+      genblock(l,n);
       if (l>=0)
       {
         ctxvar *v=ctxstack+getvarerr(l);
@@ -628,7 +697,7 @@ void genblock(int n, int p)
     {
       int xx=tree[x].x;
       int xy=tree[tree[x].y].x;
-      genblockaffmatrixl(n,xx,xy,y); 
+      genblockaffmatrixl(n,xx,xy,y);
     }
     else
     {
@@ -668,12 +737,21 @@ void genblock(int n, int p)
       nn=newcoeff(x,tree[y].x,tree[y].y);
       tree[n]=tree[nn];
     }
-    break;	
+    break;
+  case Fcall:
+    {
+      int seq = newcall("_(_)",addlistleft(x,y));
+      if (FC_call<0) die(n,"closures not available in this version");
+      tree[n]=tree[seq];
+      genblock(n,p);
+    }
+    break;
   case Frefarg:
     x=detag(x);
     if (tree[x].f==Ffacteurmat)
     {
       int l=getlvalue(x);
+      genblock(l,n);
       if (l>=0)
       {
         ctxvar *v=ctxstack+getvarerr(l);
@@ -682,24 +760,58 @@ void genblock(int n, int p)
     }
     genblock(x,n);
     break;
-  case Fentry:/*may be an hidden function call*/
-    if (findfunction(entryname(n))==-1) /*it is not a function call*/
+  case Fentry:
     {
-      if (getvar(n)==-1)
-      {/*The variable has not been declared.
-        * We declare it as a function-local variable initialized
-        * to 'var. The global flag is set to avoid being block-local.
-        * It is discarded in genblockdeffunc. This can be changed.
-        */
-        int iv=newquotenode(entryname(n),-1);
-        pushvar(n,(1<<Cuser)|(1<<Cglobal)|(1<<Cundeclared),Ggen,iv);
-        if (warn) 
-          warning(n,"variable undeclared");
+      int nf, nv = getvar(n);
+      if (nv>=0) break;
+      nf = findfunction(entryname(n));
+      if (nf<0) /*Undeclared variable*/
+      {
+        gpfunc *gp = lfunc+currfunc;
+        userfunc *uf = gp->user;
+        if (uf->flag&(1<<UFclosure))
+        {
+          /* We assume the closure contains the variable */
+          pushvar(n,(1<<Cuser)|(1<<Carg)|(1<<Ccompo)|(1<<Cclosed),Ggen,-1);
+          uf->narg++;
+        }
+        else
+        {/*The variable has not been declared.
+          * We declare it as a function-local variable initialized
+          * to 'var. The global flag is set to avoid being block-local.
+          * It is discarded in genblockdeffunc. This can be changed.
+          */
+          int iv=newquotenode(entryname(n),-1);
+          pushvar(n,(1<<Cuser)|(1<<Cglobal)|(1<<Cundeclared),Ggen,iv);
+          if (warn)
+            warning(n,"variable undeclared");
+        }
+        break;
+      } else
+      {
+        /* This is a closure */
+        gpfunc *gp = lfunc+nf;
+        const char * code = gp->proto.code;
+        GPspec spec = gp->spec;
+        if (spec==GPuser && gp->user->flag&(1<<UFclosure))
+        {
+          int z=newcall("_closure",genclosedlist(nf,newstringnode(lfunc[nf].gpname,-1)));
+          tree[n]=tree[z];
+          genblock(n, p);
+          break;
+        }
+        else if (spec <= 0 && (spec!=GPpari || (code && has_mand_arg(code))))
+        {
+          int z=newcall("_closure",newstringnode(lfunc[nf].gpname,-1));
+          tree[n]=tree[z];
+          genblock(n, p);
+          break;
+        }
       }
-      break;
-    }/*else it is a function call*/
-    tree[n].f=Fentryfunc;
-  case Fentryfunc:/*fall through*/
+      /*else it is a function call*/
+      tree[n].f=Ffunction;
+    }
+  case Ffunction:/*fall through*/
     if (x==OPn)
     {
       if (tree[y].f==Fsmall)
@@ -725,6 +837,7 @@ void genblock(int n, int p)
       {
         int z=newgetgvec("t_COL");
         tree[n]=tree[z];
+        break;
       }
       else
       {
@@ -743,7 +856,9 @@ void genblock(int n, int p)
         dy=detag(tree[y].x);
       if (tree[dy].f==Ffacteurmat)
       {
-        ctxvar *v = ctxstack+getvarerr(getlvaluerr(dy));
+        int l=getlvalue(dy);
+        genblock(l,n);
+        ctxvar *v = ctxstack+getvarerr(l);
         v->flag|=1<<Ccompo;
       }
       multi++;
@@ -753,7 +868,22 @@ void genblock(int n, int p)
     else
     {
       int nf=findfunction(value[x].val.str);
-      if (nf>=0 && lfunc[nf].spec>0)
+      gpfunc *gp = lfunc+currfunc;
+      userfunc *uf = gp->user;
+      int nv = getvar(n);
+      if (nf<0 && uf->flag&(1<<UFclosure) && nv==-1)
+      {
+        pushvar(n,(1<<Cuser)|(1<<Carg)|(1<<Ccompo)|(1<<Cclosed),Ggen,-1);
+        uf->narg++;
+        genblock(y,n);
+      }
+      else if (nf<0 && nv>=0)
+      {
+        int seq = newnode(Fcall,newnode(Fentry,x,-1),y);
+        tree[n]=tree[seq];
+        genblock(n,p);
+      }
+      else if (nf>=0 && lfunc[nf].spec>0)
         genblockfuncspec(n,p,lfunc+nf);
       else
       {
@@ -765,12 +895,12 @@ void genblock(int n, int p)
     break;
   case Fdeffunc:
     if (currfunc!=-1)
-      die(n,"nested function definition not implemented");
+      die(n,"Internal error : nested function");
     genblockdeffunc(n);
     break;
   case Fconst:
   case Fsmall:
-  case Fgnil:
+  case Fnoarg:
     break;
   case Ftag:
     genblock(x,n);
